"""
rich-codex is a minimal Python module for generating terminal screenshots from shell commands and text snippets.

It should work as a standalone command-line tool, however it is primarily intended
for use with the rich-codex GitHub Action.
"""

__version__ = "1.2.0"

from rich_codex import rich_img  # noqa: F401



import logging
from pathlib import Path

from git import Repo
from git.exc import InvalidGitRepositoryError
from jsonschema import Draft4Validator
from jsonschema.exceptions import ValidationError

log = logging.getLogger("rich-codex")


def clean_images(clean_img_paths_raw, img_obj, codex_obj):
    """Delete any images matching CLEAN_IMG_PATHS that were not generated.

    Useful to remove existing files when a target filename is changed.
    """
    clean_img_patterns = clean_img_paths_raw.splitlines() if clean_img_paths_raw else []

    if len(clean_img_patterns) == 0:
        log.debug("[dim]Nothing found to clean in 'clean_img_paths'")
        return []

    # Search glob patterns for images
    all_img_paths = set()
    for pattern in clean_img_patterns:
        for matched_path in Path.cwd().glob(pattern):
            all_img_paths.add(matched_path.resolve())
    if len(all_img_paths) == 0:
        log.debug("[dim]No files found matching 'clean_img_paths' glob patterns")
        return []

    # Collect list of generated images
    known_img_paths = set()
    if img_obj:
        for img_path in img_obj.img_paths:
            known_img_paths.add(Path(img_path).resolve())
    if codex_obj:
        for img in codex_obj.rich_imgs:
            for img_path in img.img_paths:
                known_img_paths.add(Path(img_path).resolve())

    # Paths found by glob that weren't generated
    clean_img_paths = all_img_paths - known_img_paths
    if len(clean_img_paths) == 0:
        log.debug("[dim]All files found matching 'clean_img_paths' were generated in this run. Nothing to clean.")
        return []

    for path in clean_img_paths:
        path_to_delete = Path(path).resolve()
        path_relative = path_to_delete.relative_to(Path.cwd())
        log.info(f"Deleting '{path_relative}'")
        path_to_delete.unlink()

    return clean_img_paths


def check_git_status():
    """Check if the working directory is a clean git repo."""
    try:
        repo = Repo(Path.cwd().resolve(), search_parent_directories=True)
        if repo.is_dirty():
            return (False, "Found uncommitted changes")
    except InvalidGitRepositoryError:
        return (False, "Does not appear to be a git repository")
    return (True, "Git repo looks good.")


def validate_config(schema, config, filename, line_number=None):
    """Validate a config file string against the rich-codex JSON schema."""
    ln_text = f"line {line_number} " if line_number else ""
    v = Draft4Validator(schema)
    if not v.is_valid(config):
        err_msg = f"[red][✗] Rich-codex config in '{filename}' {ln_text}was invalid"

        for error in sorted(v.iter_errors(config), key=str):
            err_msg += f"\n - {error.message}"
            if len(error.context):
                err_msg += ":"
            for suberror in sorted(error.context, key=lambda e: e.schema_path):
                err_msg += f"\n     * {suberror.message}"
        raise ValidationError(err_msg, v)



import logging
from datetime import datetime
from os import getenv
from sys import exit

from jsonschema.exceptions import ValidationError
from rich.console import Console
from rich.logging import RichHandler

from rich_codex import __version__, codex_search, rich_img, utils

import rich_click as click

click.rich_click.OPTION_ENVVAR_FIRST = True
click.rich_click.ENVVAR_STRING = "[env: {}]"

log = logging.getLogger()


@click.command()
@click.option(
    "--search-include",
    envvar="SEARCH_INCLUDE",
    show_envvar=True,
    help="Glob patterns to search for rich-codex comments",
)
@click.option(
    "--search-exclude",
    envvar="SEARCH_EXCLUDE",
    show_envvar=True,
    help="Glob patterns to exclude from search for rich-codex comments",
)
@click.option(
    "--no-search",
    is_flag=True,
    envvar="NO_SEARCH",
    show_envvar=True,
    help="Set to disable searching for rich-codex comments",
)
@click.option(
    "--command",
    envvar="COMMAND",
    show_envvar=True,
    help="Specify a command to run to capture output",
)
@click.option(
    "--timeout",
    type=int,
    default=5,
    envvar="TIMEOUT",
    show_envvar=True,
    show_default=True,
    help="Maximum run time for command (seconds)",
)
@click.option(
    "--before-command",
    envvar="BEFORE_COMMAND",
    show_envvar=True,
    help="Setup commands to run before running main output command",
)
@click.option(
    "--after-command",
    envvar="AFTER_COMMAND",
    show_envvar=True,
    help="Setup commands to run after running main output command",
)
@click.option(
    "--snippet",
    envvar="SNIPPET",
    show_envvar=True,
    help="Literal code snippet to render",
)
@click.option(
    "--snippet-syntax",
    envvar="SNIPPET_SYNTAX",
    show_envvar=True,
    help="Language to use for snippet sytax highlighting",
)
@click.option(
    "--img-paths",
    envvar="IMG_PATHS",
    show_envvar=True,
    help="Path to image filenames if using 'command' or 'snippet'",
)
@click.option(
    "--clean-img-paths",
    envvar="CLEAN_IMG_PATHS",
    show_envvar=True,
    help="Remove any matching files that are not generated",
)
@click.option(
    "--configs",
    envvar="RC_CONFIGS",
    show_envvar=True,
    help="Paths to YAML config files",
)
@click.option(
    "--fake-command",
    envvar="FAKE_COMMAND",
    show_envvar=True,
    help="Pretend command to show in the screenshot prompt",
)
@click.option(
    "--hide-command",
    is_flag=True,
    envvar="HIDE_COMMAND",
    show_envvar=True,
    help="Hide the terminal prompt with the command at the top of the output",
)
@click.option(
    "--title-command",
    is_flag=True,
    envvar="TITLE_COMMAND",
    show_envvar=True,
    help="Use the command as the terminal title if not set explicitly",
)
@click.option(
    "--head",
    type=int,
    envvar="RC_HEAD",
    show_envvar=True,
    help="Show only the first N lines of output",
)
@click.option(
    "--tail",
    type=int,
    envvar="RC_TAIL",
    show_envvar=True,
    help="Show only the last N lines of output",
)
@click.option(
    "--trim-after",
    envvar="TRIM_AFTER",
    show_envvar=True,
    help="Don't print any more lines after this string is found",
)
@click.option(
    "--truncated-text",
    default="[..truncated..]",
    envvar="TRUNCATED_TEXT",
    show_envvar=True,
    help="Text to show when --head or --tail truncate content",
)
@click.option(
    "--skip-git-checks",
    is_flag=True,
    envvar="SKIP_GIT_CHECKS",
    show_envvar=True,
    help="Skip safety checks for git repos",
)
@click.option(
    "--no-confirm",
    is_flag=True,
    envvar="NO_CONFIRM",
    show_envvar=True,
    help="Set to skip confirmation prompt before running commands",
)
@click.option(
    "--min-pct-diff",
    envvar="MIN_PCT_DIFF",
    type=float,
    default=0,
    show_envvar=True,
    show_default=True,
    help="Minimum file percentage change required to update image",
)
@click.option(
    "--skip-change-regex",
    envvar="SKIP_CHANGE_REGEX",
    show_envvar=True,
    help="Skip image update if file changes match regex",
)
@click.option(
    "--terminal-width",
    envvar="TERMINAL_WIDTH",
    show_envvar=True,
    help="Width of the terminal",
)
@click.option(
    "--terminal-min-width",
    type=int,
    default=80,
    envvar="TERMINAL_MIN_WIDTH",
    show_envvar=True,
    show_default=True,
    help="Minimum width of the terminal when trimming",
)
@click.option(
    "--notrim",
    is_flag=True,
    envvar="NOTRIM",
    show_envvar=True,
    help="Disable automatic trimming of terminal width",
)
@click.option(
    "--terminal-theme",
    envvar="TERMINAL_THEME",
    show_envvar=True,
    help="Colour theme",
)
@click.option(
    "--snippet-theme",
    envvar="SNIPPET_THEME",
    show_envvar=True,
    help="Snippet Pygments theme",
)
@click.option(
    "--use-pty",
    is_flag=True,
    envvar="USE_PTY",
    show_envvar=True,
    help="Use a pseudo-terminal for commands (may capture coloured output)",
)
@click.option(
    "--created-files",
    envvar="CREATED_FILES",
    show_envvar=True,
    help="Save a list of created files to this file",
    metavar="<filename>",
)
@click.option(
    "--deleted-files",
    envvar="DELETED_FILES",
    show_envvar=True,
    help="Save a list of deleted files to this file",
    metavar="<filename>",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    envvar="LOG_VERBOSE",
    show_envvar=True,
    help="Print verbose output to the console.",
)
@click.option(
    "--save-log",
    is_flag=True,
    envvar="LOG_SAVE",
    show_envvar=True,
    help="Save a verbose log to a file (automatic filename).",
    metavar="<filename>",
)
@click.option(
    "-l",
    "--log-file",
    envvar="LOG_FILENAME",
    show_envvar=True,
    help="Save a verbose log to a file (specific filename).",
    metavar="<filename>",
)
def main(
    search_include,
    search_exclude,
    no_search,
    command,
    timeout,
    before_command,
    after_command,
    snippet,
    snippet_syntax,
    img_paths,
    clean_img_paths,
    configs,
    fake_command,
    hide_command,
    title_command,
    head,
    tail,
    trim_after,
    truncated_text,
    skip_git_checks,
    no_confirm,
    min_pct_diff,
    skip_change_regex,
    terminal_width,
    terminal_min_width,
    notrim,
    terminal_theme,
    snippet_theme,
    use_pty,
    created_files,
    deleted_files,
    verbose,
    save_log,
    log_file,
):
    """Create rich code images for your docs."""
    # Sensible defaults
    no_confirm = True if not no_confirm and getenv("GITHUB_ACTIONS") else no_confirm
    force_terminal = True if getenv("GITHUB_ACTIONS") or getenv("FORCE_COLOR") or getenv("PY_COLORS") else None
    terminal_width = int(terminal_width) if type(terminal_width) is str else terminal_width
    terminal_min_width = int(terminal_min_width) if type(terminal_min_width) is str else terminal_min_width
    saved_image_paths = []
    cleaned_paths = []
    num_skipped_images = 0
    num_saved_images = 0
    img_obj = None
    codex_obj = None

    # Set up the logger
    log.setLevel(logging.DEBUG)

    # Set up logs to the console
    log.addHandler(
        RichHandler(
            level=logging.DEBUG if verbose else logging.INFO,
            console=Console(
                stderr=True,
                force_terminal=force_terminal,
            ),
            show_time=False,
            markup=True,
            rich_tracebacks=True,
            show_path=False,
            tracebacks_suppress=[click],
        )
    )

    # Set up logs to a file if we asked for one
    if save_log and not log_file:
        timestamp = datetime.now().strftime("%Y.%m.%d--%H.%M.%S.%f")
        log_file = f"rich_codex_{timestamp}.log"

    if log_file:
        log_fh = logging.FileHandler(log_file, encoding="utf-8")
        log_fh.setLevel(logging.DEBUG)
        log_fh.setFormatter(logging.Formatter("[%(asctime)s] %(name)-20s [%(levelname)-7s]  %(message)s"))
        log.addHandler(log_fh)

    log.info(f"[bold]rich-codex[/] ⚡️📖⚡️ [dim]version {__version__}[/dim]")

    # Check git status
    git_status, git_status_msg = utils.check_git_status()
    if skip_git_checks or git_status:
        log.debug(f"Git status check: {git_status_msg} (skip_git_checks: {skip_git_checks})")
    elif not git_status:
        log.error(f"[bright_red]Error with git:[/] [red]{git_status_msg}")
        log.info("Please resolve and run again, or use '--skip-git-checks'")
        exit(1)

    if no_confirm:
        log.debug("Skipping confirmation of commands")
    if force_terminal:
        log.debug("Forcing terminal logging output")
    if terminal_width:
        log.info(f"Setting terminal width to {terminal_width}")
    if terminal_min_width and not notrim:
        log.info(f"Trimming terminal output down to a minimum of {terminal_min_width}")
    if terminal_width and terminal_min_width:
        if terminal_min_width > terminal_width:
            log.error(
                "terminal_min_width ({}) > terminal_width ({})! Disabling terminal_min_width".format(
                    terminal_min_width, terminal_width
                )
            )
            terminal_min_width = None

    # Console for printing to stdout
    console = Console(
        force_terminal=force_terminal,
        highlight=False,
        width=100 if getenv("GITHUB_ACTIONS") else None,
    )

    # Check for mutually exclusive options
    if command and snippet:
        raise click.BadOptionUsage("--command", "Please use either --command OR --snippet but not both")
    if (command or snippet) and not img_paths:
        raise click.BadOptionUsage("--img-paths", "--img-paths is required when using --command or --snippet")

    # Generate image from a supplied command / snippet
    if command or snippet:
        img_obj = rich_img.RichImg(
            snippet_syntax=snippet_syntax,
            timeout=timeout,
            before_command=before_command,
            after_command=after_command,
            fake_command=fake_command,
            hide_command=hide_command,
            title_command=title_command,
            head=head,
            tail=tail,
            trim_after=trim_after,
            truncated_text=truncated_text,
            min_pct_diff=min_pct_diff,
            skip_change_regex=skip_change_regex,
            terminal_width=terminal_width,
            terminal_min_width=terminal_min_width,
            notrim=notrim,
            terminal_theme=terminal_theme,
            snippet_theme=snippet_theme,
            use_pty=use_pty,
            console=console,
        )
        img_obj.no_confirm = no_confirm
        if command:
            log.info(f"Command: [white on black] {command} [/]")
            img_obj.command = command
        if snippet:
            log_snippet = snippet[0:30].replace("\n", " ")
            log.info(f"Snippet: [white on black] {log_snippet}... [/]")
            img_obj.snippet = snippet
        img_obj.img_paths = img_paths.splitlines() if img_paths else []
        if img_obj.confirm_command():
            img_obj.get_output()
            img_obj.save_images()
            saved_image_paths += img_obj.saved_img_paths
            num_saved_images += img_obj.num_img_saved
            num_skipped_images += img_obj.num_img_skipped

    # Generate images from config files

    # Search files for codex strings
    codex_obj = codex_search.CodexSearch(
        search_include=search_include,
        search_exclude=search_exclude,
        configs=configs,
        no_confirm=no_confirm,
        snippet_syntax=snippet_syntax,
        timeout=timeout,
        before_command=before_command,
        after_command=after_command,
        hide_command=hide_command,
        title_command=title_command,
        head=head,
        tail=tail,
        trim_after=trim_after,
        truncated_text=truncated_text,
        min_pct_diff=min_pct_diff,
        skip_change_regex=skip_change_regex,
        terminal_width=terminal_width,
        terminal_min_width=terminal_min_width,
        notrim=notrim,
        terminal_theme=terminal_theme,
        snippet_theme=snippet_theme,
        use_pty=use_pty,
        console=console,
    )
    try:
        codex_obj.parse_configs()
    except ValidationError as e:
        log.critical(e)
        exit(1)
    if no_search:
        log.info("Skipping file search")
    else:
        num_errors = codex_obj.search_files()
        if num_errors > 0:
            log.error("Found errors whilst running")
            exit(1)
    codex_obj.collapse_duplicates()
    codex_obj.confirm_commands()
    codex_obj.check_duplicate_paths()
    codex_obj.save_all_images()
    saved_image_paths += codex_obj.saved_img_paths
    num_saved_images += codex_obj.num_img_saved
    num_skipped_images += codex_obj.num_img_skipped

    # Clean unrecognised images
    if clean_img_paths:
        cleaned_paths = utils.clean_images(clean_img_paths, img_obj, codex_obj)

    # Write saved file paths to disk
    if created_files and len(saved_image_paths):
        log.info(f"Saving list of new file paths to: [magenta]{created_files}[/]")
        with open(created_files, "w") as f:
            f.write("\n".join(saved_image_paths))

    # Write cleaned file paths to disk
    if deleted_files and len(cleaned_paths):
        log.info(f"Saving list of deleted file paths to: [magenta]{deleted_files}[/]")
        with open(deleted_files, "w") as f:
            f.write("\n".join([str(path) for path in cleaned_paths]))

    if num_skipped_images > 0:
        log.info(f"[dim]Skipped {num_skipped_images} images 🤫")
    if num_saved_images > 0:
        log.info(f"Saved {num_saved_images} images ✨")
    if len(cleaned_paths) > 0:
        log.info(f"Deleted {len(cleaned_paths)} images 💥")
    if num_skipped_images == 0 and num_saved_images == 0:
        log.warning("Couldn't find anything to do 🙄")


if __name__ == "__main__":
    main()



import difflib
import json
import logging
import os
import re
import signal
import subprocess
from pathlib import Path
from shutil import copyfile
from tempfile import gettempdir, mkstemp

import rich.terminal_theme
import yaml
from Levenshtein import ratio
from rich.ansi import AnsiDecoder
from rich.console import Console
from rich.prompt import Confirm
from rich.syntax import Syntax

log = logging.getLogger("rich-codex")

# Parse the config schema file to get config attributes
config_schema_fn = Path(__file__).parent / "config-schema.yml"
with config_schema_fn.open() as fh:
    config_schema = yaml.safe_load(fh)
RICH_IMG_ATTRS = config_schema["properties"]["outputs"]["items"]["properties"].keys()

# Base list of commands to ignore
IGNORE_COMMANDS = ["rm", "cp", "mv", "sudo"]

# Base list of diff regexes to ignore, split up by filetype suffix
IGNORE_REGEXES = {".pdf": [r"/CreationDate"]}


class RichImg:
    """Image generation for rich-codex.

    Objects from this class are typically used once per screenshot.
    """

    def __init__(
        self,
        command=None,
        working_dir=None,
        snippet=None,
        img_paths=None,
        snippet_syntax=None,
        timeout=5,
        before_command=None,
        after_command=None,
        title=None,
        fake_command=None,
        hide_command=False,
        title_command=False,
        extra_env=None,
        head=None,
        tail=None,
        trim_after=None,
        truncated_text="[..truncated..]",
        min_pct_diff=0,
        skip_change_regex=None,
        terminal_width=None,
        terminal_min_width=80,
        notrim=False,
        terminal_theme=None,
        snippet_theme=None,
        use_pty=False,
        console=None,
        source_type=None,
        source=None,
    ):
        """Initialise the RichImg object with core console options."""
        self.command = command
        self.working_dir = Path.cwd() if working_dir is None else working_dir
        self.snippet = snippet
        self.img_paths = [] if img_paths is None else img_paths
        self.snippet_syntax = snippet_syntax
        self.timeout = timeout
        self.before_command = before_command
        self.after_command = after_command
        self.title = "" if title is None else title
        self.fake_command = fake_command
        self.hide_command = hide_command
        self.title_command = title_command
        self.extra_env = {} if extra_env is None else extra_env
        self.head = None if head is None else int(head)
        self.tail = None if tail is None else int(tail)
        self.trim_after = trim_after
        self.truncated_text = truncated_text
        self.min_pct_diff = min_pct_diff
        self.skip_change_regex = skip_change_regex
        self.terminal_width = None if terminal_width is None else int(terminal_width)
        self.terminal_min_width = None if terminal_min_width is None else int(terminal_min_width)
        if self.terminal_width and self.terminal_min_width and self.terminal_min_width > self.terminal_width:
            self.terminal_min_width = None
        self.notrim = notrim
        self.terminal_theme = terminal_theme
        self.snippet_theme = snippet_theme
        self.use_pty = use_pty
        self.console = Console() if console is None else console
        self.capture_console = None
        self.saved_img_paths = []
        self.num_img_saved = 0
        self.num_img_skipped = 0
        self.no_confirm = False
        self.aborted = False
        self.source_type = source_type
        self.source = source

    def __eq__(self, other):
        """Compare RichImg objects for equality."""
        if not isinstance(other, RichImg):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return all(getattr(self, attr) == getattr(other, attr) for attr in RICH_IMG_ATTRS)

    def __hash__(self):
        """Hash stable identifier of RichImg object based on important attributes."""
        attrs = str([getattr(self, attr) for attr in RICH_IMG_ATTRS])
        return hash(attrs)

    def _hash_no_fn(self):
        """Hash stable identifier of RichImg object based without output filenames."""
        attrs = str([getattr(self, attr) for attr in RICH_IMG_ATTRS if attr != "img_paths"])
        return hash(attrs)

    def confirm_command(self):
        """Prompt user to confirm running command."""
        if self.command is None or self.no_confirm:
            return True
        return Confirm.ask(f"Command: [white on black] {self.command} [/] Run?", console=self.console)

    def run_command(self):
        """Capture output from a supplied command and save to an image."""
        if self.command is None:
            log.debug("Tried to generate image with no command")
            return

        self.command = self.command.strip()

        for ignore in IGNORE_COMMANDS:
            if any(cmd_part.strip().startswith(ignore) for cmd_part in self.command.split("&;")):
                log.warning(f"Ignoring command because it contained '{ignore}': [white on black] {self.command} [/]")
                self.aborted = True
                return False

        if self.title == "" and self.title_command:
            self.title = self.fake_command if self.fake_command else self.command

        if self.use_pty:
            log.debug(f"Running command '{self.command}' with pty")

            try:
                import fcntl
                import pty
                import struct
                import termios

                run_with_pty = True
            except ImportError:
                # fallback method needed
                log.warning(
                    "Could not use pty, import failed (are you using Windows? pty is not usable there). "
                    "Falling back to subprocess."
                )
                run_with_pty = False
        else:
            log.debug(f"Running command '{self.command}' with subprocess")
            run_with_pty = False

        # Create working directory if it doesn't already exist
        self.working_dir.mkdir(parents=True, exist_ok=True)

        # Set up the command environment vars, with extra_env if set
        command_env = os.environ.copy()
        if len(self.extra_env):
            log.debug(f"Adding extra env variables: {self.extra_env}")
            command_env.update({str(k): str(v) for k, v in self.extra_env.items()})

        # Run before_command if set
        if self.before_command:
            log.debug("Running 'before_command'")
            # TODO: Can't get inspect() to output to a log call
            # NOTE: https://github.com/Textualize/rich/discussions/2378
            # log.debug(
            #     inspect(
            subprocess.run(
                self.before_command,
                cwd=self.working_dir,
                shell=True,
                env=command_env,
                capture_output=True,
            )
            #         , title="'before_command' results",
            #         docs=False,
            #     )
            # )

        # Run the command with a fake tty to try to get colours
        if run_with_pty:
            read_end, write_end = pty.openpty()

            # Resize routine for pty
            # First, get our own current terminal size
            # (struct is documented here: https://www.delorie.com/djgpp/doc/libc/libc_495.html)
            size = fcntl.ioctl(0, termios.TIOCGWINSZ, struct.pack("HHHH", 0, 0, 0, 0))

            # Rewrite size with selected terminal width if set
            if self.terminal_width is not None:
                data = struct.unpack("HHHH", size)
                size = struct.pack("HHHH", data[0], self.terminal_width, 0, 0)

            # Issue command to pty to resize
            fcntl.ioctl(write_end, termios.TIOCSWINSZ, size)
            fcntl.ioctl(read_end, termios.TIOCSWINSZ, size)
            signal.signal(signal.SIGWINCH, lambda s, f: fcntl.ioctl(write_end, termios.TIOCSWINSZ, size))
            signal.signal(signal.SIGWINCH, lambda s, f: fcntl.ioctl(read_end, termios.TIOCSWINSZ, size))

            # Run subprocess in pty
            try:
                process = subprocess.Popen(
                    self.command,
                    cwd=self.working_dir,
                    shell=True,
                    env=command_env,
                    close_fds=True,
                    preexec_fn=os.setsid,
                    stdin=write_end,
                    stdout=write_end,
                    stderr=write_end,
                )
                process.wait(timeout=float(self.timeout))
            except subprocess.TimeoutExpired:
                log.info(f"Command '{self.command}' timed out after {self.timeout} seconds")
                os.killpg(os.getpgid(process.pid), signal.SIGTERM)
            os.close(write_end)

            output_arr = []

            # This loop will keep going until no more data is incoming (child process closed their pipe)
            while True:
                try:
                    data = os.read(read_end, 1024)
                except OSError:
                    data = b""

                if data:
                    output_arr.append(data)
                else:
                    break

            os.close(read_end)
            output = b"".join(output_arr).decode("utf-8")

        # Run the command without messing with ttys
        else:
            try:
                process = subprocess.Popen(
                    self.command,
                    cwd=self.working_dir,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    shell=True,  # Needed for pipes
                    env=command_env,
                    start_new_session=True,  # Needed for subprocess termination
                )
                output, errs = process.communicate(timeout=float(self.timeout))
            except subprocess.TimeoutExpired:
                log.info(f"Command '{self.command}' timed out after {self.timeout} seconds")
                os.killpg(os.getpgid(process.pid), signal.SIGKILL)
                output, errs = process.communicate()
            output = output.decode("utf-8")

        # Run after_command if set
        if self.after_command:
            log.debug("Running 'after_command'")
            # TODO: Can't get inspect() to output to a log call
            # NOTE: https://github.com/Textualize/rich/discussions/2378
            # log.debug(
            #     inspect(
            subprocess.run(
                self.after_command,
                cwd=self.working_dir,
                shell=True,
                env=command_env,
                capture_output=True,
            )
            #         , title="'after_command' results",
            #         docs=False,
            #     )
            # )

        decoder = AnsiDecoder()

        # Count lines and find longest line
        print_lines = []
        max_line_length = 0
        for line in decoder.decode(output):
            print_lines.append(True)
            max_line_length = max(len(line), max_line_length)

        # If terminal_min_width is set, find longest line
        t_width = int(self.terminal_width) if self.terminal_width else None
        if not self.notrim and self.terminal_min_width:
            t_width = int(self.terminal_min_width)
            t_width = max(t_width, max_line_length)
            log.debug(f"Setting terminal width to {t_width}")
        self.capture_console = Console(
            file=open(os.devnull, "w"),
            force_terminal=True,
            color_system="truecolor",
            highlight=False,
            record=True,
            width=t_width,
        )

        # Set head / tail print set
        if self.head and self.head >= len(print_lines):
            self.head = None
        if self.tail and self.tail >= len(print_lines):
            self.tail = None
        if self.head is not None or self.tail is not None:
            print_lines = [False] * len(print_lines)
            if self.head is not None:
                print_lines[: self.head] = [True] * self.head
            if self.tail is not None:
                print_lines[len(print_lines) - self.tail :] = [True] * self.tail

        # Print the command
        if not self.hide_command:
            self.capture_console.print(
                "$ {}".format(
                    self.fake_command if self.fake_command else self.command,
                )
            )

        # Decode and print the output (captured)
        for idx, line in enumerate(decoder.decode(output)):
            if print_lines[idx]:
                self.capture_console.print(line)
                # Trim text after trim_after
                if self.trim_after and self.trim_after in line:
                    break
            elif (self.head is not None or self.tail is not None) and self.truncated_text:
                self.capture_console.print(self.truncated_text, style="italic dim")
                self.truncated_text = None

    def format_snippet(self):
        """Take a text snippet and format it using rich."""
        if self.snippet is None:
            log.debug("Tried to format snippet with no snippet")
            return

        # Reformat JSON with pretty printing, because we can
        if self.snippet_syntax == "json" or self.snippet_syntax is None:
            try:
                json_snippet = json.loads(self.snippet)
                self.snippet = json.dumps(json_snippet, indent=4, sort_keys=True)
                self.snippet_syntax = "json"
            except json.decoder.JSONDecodeError:
                pass

        # Adjust terminal width if min-width set
        t_width = int(self.terminal_width) if self.terminal_width else None
        if not self.notrim and self.terminal_min_width:
            t_width = int(self.terminal_min_width)
            for line in self.snippet.splitlines():
                t_width = max(len(line), t_width)
            log.debug(f"Setting terminal width to {t_width}")

        self.capture_console = Console(
            file=open(os.devnull, "w"),
            force_terminal=True,
            color_system="truecolor",
            highlight=False,
            record=True,
            width=t_width,
        )

        # Print with rich Syntax highlighter
        log.debug(f"Formatting snippet as {self.snippet_syntax}")
        snippet_theme = "monokai" if self.snippet_theme is None else self.snippet_theme  # Same default as Rich
        syntax = Syntax(self.snippet, self.snippet_syntax, theme=snippet_theme)
        self.capture_console.print(syntax)

    def get_output(self):
        """Either run command or format snippet, depending on what is set."""
        if self.command is not None:
            self.run_command()
        elif self.snippet is not None:
            self.format_snippet()
        else:
            log.warning("Tried to get output with no command or snippet")

    def _enough_image_difference(self, new_fn, old_fn):
        new_file = Path(new_fn)
        old_file = Path(old_fn)
        create_file = True
        log_msg = ""

        # Target file doesn't exist yet, nothing to compare against
        if not old_file.exists():
            log_msg = "new image"

        else:
            # Percentage change in file
            # This method works even with entirely binary files, no decoding required
            pct_change = (1 - ratio(new_file.read_bytes(), old_file.read_bytes())) * 100.0
            if pct_change <= float(self.min_pct_diff):
                create_file = False
            log_msg = f"{pct_change:.2f}% change"

            # No point in looking for a diff if the files are identical
            if pct_change > 0:

                # Regex on file diff to skip
                skip_regexes = list(r for r in IGNORE_REGEXES.get(new_file.suffix, []))  # deep copy
                if self.skip_change_regex:
                    skip_regexes.extend(self.skip_change_regex.splitlines())
                if len(skip_regexes) > 0:
                    new_file_lines = new_file.read_text(errors="ignore").splitlines()
                    old_file_lines = old_file.read_text(errors="ignore").splitlines()
                    log.info("Checking diff")

                    # Only continue if we found something to diff with
                    if len(new_file_lines) > 0 or len(old_file_lines) > 0:
                        log.info("starting..")
                        diffs = difflib.Differ().compare(new_file_lines, old_file_lines)
                        log.info("generated diff ")
                        log.info(f"len: {len(list(diffs))}")
                        lost_lines = [d for d in diffs if d.startswith("-")]

                        # Only continue if there was some diff
                        log.info(f"Found {lost_lines} lines")
                        if len(lost_lines) > 0:
                            matched_lost_lines = []
                            for skip_regex in skip_regexes:
                                for line in lost_lines:
                                    if re.search(skip_regex, line):
                                        matched_lost_lines.append(line)

                            log_msg += f", {len(matched_lost_lines)}/{len(lost_lines)} diff lines matched regex filters"
                            if len(matched_lost_lines) == len(lost_lines):
                                create_file = False
                        else:
                            log_msg += ", no diff found"
                    else:
                        log_msg += ", no text to diff"

        old_fn_relative = old_file.resolve().relative_to(Path.cwd())
        if create_file:
            self.num_img_saved += 1
            if old_fn in self.saved_img_paths:
                log.warning(f"More than one image with file name '{old_fn}'")
            else:
                self.saved_img_paths.append(old_fn)
            log.info(f"Saved: '{old_fn_relative}' ({log_msg})")
        else:
            self.num_img_skipped += 1
            log.debug(f"[dim]Skipped: '{old_fn_relative}' ({log_msg})")

        return create_file

    def save_images(self):
        """Save the images to the specified filenames."""
        if self.aborted:
            return
        if len(self.img_paths) == 0:
            log.warning("Tried to save images with no paths")
            return

        # Set up theme
        terminal_theme = None
        if self.terminal_theme:
            try:
                terminal_theme = getattr(rich.terminal_theme, self.terminal_theme)
            except AttributeError:
                log.error(
                    "[red]Theme '{}' not found![/] Falling back to default for [magenta]{}".format(
                        self.terminal_theme, ", ".join(self.img_paths)
                    )
                )

        # Save image as requested with $IMG_PATHS
        svg_img = None
        png_img = None
        pdf_img = None
        for filename in self.img_paths:

            # Make directories if necessary
            Path(filename).parent.mkdir(parents=True, exist_ok=True)

            # If already made this image, copy it from the last destination
            if filename.lower().endswith(".png") and png_img is not None:
                log.debug(f"Using '{png_img}' for '{filename}'")
                if self._enough_image_difference(png_img, filename):
                    copyfile(png_img, filename)
                continue
            if filename.lower().endswith(".pdf") and pdf_img is not None:
                log.debug(f"Using '{pdf_img}' for '{filename}'")
                if self._enough_image_difference(pdf_img, filename):
                    copyfile(pdf_img, filename)
                continue
            if filename.lower().endswith(".svg") and svg_img is not None:
                log.debug(f"Using '{svg_img}' for '{filename}'")
                if self._enough_image_difference(svg_img, filename):
                    copyfile(svg_img, filename)
                continue

            # Set filenames
            tmp_filename = mkstemp()[1]

            # We always generate an SVG first
            if svg_img is None:
                svg_tmp_filename = mkstemp()[1]
                self.capture_console.save_svg(
                    svg_tmp_filename,
                    title=self.title,
                    theme=terminal_theme,
                )
            else:
                # Use already-generated SVG
                svg_tmp_filename = svg_img

            # Save the SVG image if requested
            if filename.lower().endswith(".svg"):
                if self._enough_image_difference(svg_tmp_filename, filename):
                    copyfile(svg_tmp_filename, filename)
                svg_img = filename

            # Lazy-load PNG / PDF libraries if needed
            if filename.lower().endswith(".png") or filename.lower().endswith(".pdf"):
                try:
                    from cairosvg import svg2pdf, svg2png
                except ImportError as e:
                    log.debug(e)
                    log.error("CairoSVG not installed, cannot convert SVG to PNG or PDF.")
                    log.info("Please install with cairo extra: 'rich-codex[cairo]'")
                    continue
                except OSError as e:
                    log.debug(e)
                    log.error(
                        "⚠️  Missing [link=https://cairosvg.org/documentation/]CairoSVG dependencies[/], "
                        "cannot convert SVG to PNG or PDF. ⚠️\n"
                        f"[red]Skipping image '{filename}'[/]"
                    )
                    continue

                # Convert to PNG if requested
                if filename.lower().endswith(".png"):
                    log.debug(f"Converting SVG '{svg_tmp_filename}' to PNG '{filename}'")
                    svg2png(
                        file_obj=open(svg_tmp_filename, "rb"),
                        write_to=tmp_filename,
                        dpi=300,
                        output_width=4000,
                    )
                    if self._enough_image_difference(tmp_filename, filename):
                        copyfile(tmp_filename, filename)
                        png_img = filename

                # Convert to PDF if requested
                if filename.lower().endswith(".pdf"):
                    log.debug(f"Converting SVG '{svg_tmp_filename}' to PDF '{filename}'")
                    svg2pdf(
                        file_obj=open(svg_tmp_filename, "rb"),
                        write_to=tmp_filename,
                    )
                    if self._enough_image_difference(tmp_filename, filename):
                        copyfile(tmp_filename, filename)
                        pdf_img = filename

            # Delete temprary files
            tmp_path = Path(tmp_filename)
            if tmp_path.is_relative_to(gettempdir()):
                tmp_path.unlink()

        # Delete temporary SVG file - after loop as can be reused
        tmp_svg_path = Path(svg_tmp_filename)
        if tmp_svg_path.is_relative_to(gettempdir()):
            tmp_svg_path.unlink()



import logging
import re
from pathlib import Path

import yaml
from jsonschema.exceptions import ValidationError
from rich import box
from rich.console import Console
from rich.prompt import Prompt
from rich.table import Table

from rich_codex import rich_img
from rich_codex.utils import validate_config

log = logging.getLogger("rich-codex")


class CodexSearch:
    """File search class for rich-codex.

    Looks through a set of source files for sets of configuration
    needed to generate screenshots.
    """

    def __init__(
        self,
        search_include,
        search_exclude,
        configs,
        no_confirm,
        snippet_syntax,
        timeout,
        before_command,
        after_command,
        hide_command,
        title_command,
        head,
        tail,
        trim_after,
        truncated_text,
        min_pct_diff,
        skip_change_regex,
        terminal_width,
        terminal_min_width,
        notrim,
        terminal_theme,
        snippet_theme,
        use_pty,
        console,
    ):
        """Initialize the search object."""
        self.search_include = ["**/*.md"] if search_include is None else self._clean_list(search_include.splitlines())
        self.search_exclude = ["**/.git*", "**/.git*/**", "**/node_modules/**"]
        if search_exclude is not None:
            self.search_exclude.extend(self._clean_list(search_exclude.splitlines()))
        self.configs = [".rich-codex.yml", ".github/rich-codex.yml", "docs/img/rich-codex.yml"]
        if configs is not None:
            self.configs.extend(self._clean_list(configs.splitlines()))
        self.no_confirm = no_confirm
        self.snippet_syntax = snippet_syntax
        self.timeout = timeout
        self.before_command = before_command
        self.after_command = after_command
        self.hide_command = hide_command
        self.title_command = title_command
        self.head = head
        self.tail = tail
        self.trim_after = trim_after
        self.truncated_text = truncated_text
        self.min_pct_diff = min_pct_diff
        self.skip_change_regex = skip_change_regex
        self.terminal_width = terminal_width
        self.terminal_min_width = terminal_min_width
        self.notrim = notrim
        self.terminal_theme = terminal_theme
        self.snippet_theme = snippet_theme
        self.use_pty = use_pty
        self.console = Console() if console is None else console
        self.rich_imgs = []
        self.saved_img_paths = []
        self.num_img_saved = 0
        self.num_img_skipped = 0
        self.class_config_attrs = [
            "snippet_syntax",
            "timeout",
            "before_command",
            "after_command",
            "hide_command",
            "title_command",
            "head",
            "tail",
            "trim_after",
            "truncated_text",
            "min_pct_diff",
            "skip_change_regex",
            "terminal_width",
            "terminal_min_width",
            "notrim",
            "terminal_theme",
            "snippet_theme",
            "use_pty",
        ]

        # Look in .gitignore to add to search_exclude
        try:
            with open(".gitignore", "r") as fh:
                log.debug("Appending contents of .gitignore to 'SEARCH_EXCLUDE'")
                self.search_exclude.extend(self._clean_list(fh.readlines()))
        except IOError:
            pass

        # Parse the config schema file
        config_schema_fn = Path(__file__).parent / "config-schema.yml"
        with config_schema_fn.open() as fh:
            self.config_schema = yaml.safe_load(fh)

    def _clean_list(self, unclean_lines):
        """Remove empty strings from a list."""
        clean_lines = []
        for line in unclean_lines:
            line = line.strip()
            if not line.startswith("#") and line:
                clean_lines.append(line)
        return clean_lines

    def _merge_local_class_attrs(self, local_config):
        """Update local config with class params.
        Only if not set locally and if not None at class level
        """
        for conf in self.class_config_attrs:
            if conf not in local_config and getattr(self, conf) is not None:
                local_config[conf] = getattr(self, conf)
        return local_config

    def search_files(self):
        """Search through a set of files for codex strings."""
        search_files = set()
        for pattern in self.search_include:
            for search_file in Path.cwd().glob(pattern):
                search_files.add(search_file.resolve())
        for pattern in self.search_exclude:
            if pattern.endswith("/"):
                pattern += "**/*"
            try:
                for exclude_file in Path.cwd().glob(pattern):
                    search_files.discard(exclude_file.resolve())
            except (ValueError, NotImplementedError):
                pass
        if len(search_files) == 0:
            log.debug("No files found to search")
        else:
            log.info(f"Searching {len(search_files)} files")

        # eg. <!-- RICH-CODEX TERMINAL_WIDTH=60 -->
        # eg. <!-- RICH-CODEX
        config_comment_re = re.compile(r"\s*<!\-\-\s*RICH-CODEX\s*(?P<config_str>.*)")

        # eg. ![`rich --help`](rich-cli-help.svg)
        img_cmd_re = re.compile(r"\s*!\[`(?P<cmd>[^`]+)`\]\((?P<img_path>.*?)(?=\"|\))(?P<title>[\"'].*[\"'])?\)")

        # eg. ![custom text](img/example.svg)
        # eg. ![](img/example-named.svg)
        img_snippet_re = re.compile(r"\s*!\[.*\]\((?P<img_path>.*?)(?=\"|\))(?P<title>[\"'].*[\"'])?\)")
        local_config = {}
        num_errors = 0
        num_commands = 0
        num_snippets = 0
        for file in search_files:
            file_rel_fn = Path(file).relative_to(Path.cwd())
            log.debug(f"Searching: [magenta]{file_rel_fn}[/]")
            with open(file, "r") as fh:
                line_number = 0
                in_config = False
                local_config_str = ""
                for line in fh:
                    line_number += 1

                    # Keep saving config if we're in a config block
                    if in_config:
                        local_config_str += line
                        if "-->" in line:
                            in_config = False
                            local_config_str = local_config_str.split("-->")[0]
                            continue

                    # Parse +  validate config yaml
                    if local_config_str != "" and not in_config:
                        try:
                            local_config = yaml.safe_load(local_config_str)
                            if not isinstance(local_config, dict):
                                raise ValidationError(
                                    f"Config YAML is not a dictionary: '{file_rel_fn}', line {line_number}"
                                )
                        except yaml.YAMLError as e:
                            log.error(f"[red][✗] Error parsing config YAML in '{file_rel_fn}' line {line_number}: {e}")
                            log.debug(f"Config block:\n{local_config_str}")
                            local_config = {}
                            num_errors += 1
                        local_config_str = ""

                    # Look for images
                    img_cmd_match = img_cmd_re.match(line)
                    img_match = img_snippet_re.match(line)
                    if (img_cmd_match or img_match) and not local_config.get("skip"):

                        # Logging string of original local config
                        local_config_logmsg = f" with config: {local_config}" if len(local_config) > 0 else ""

                        # Get the command and title from a command regex match
                        if img_cmd_match:
                            m = img_cmd_match.groupdict()
                            local_config["command"] = m["cmd"]
                            # Save the title if set
                            if m["title"]:
                                local_config["title"] = m["title"].strip("'\" ")
                        else:
                            m = img_match.groupdict()

                        # Counters for commands / snippets
                        if "command" in local_config:
                            num_commands += 1
                            img_type = "[blue]command[/]"
                        elif local_config.get("snippet", "") != "":
                            num_snippets += 1
                            img_type = "[red]snippet[/]"
                        # Just a regular image with no command / snippet - carry on
                        else:
                            log.debug(f"[dim]Skipped markdown image, line {line_number}: {m}")
                            if len(local_config) > 0:
                                log.warn(f"Skipped image but local_config was not empty: {local_config}")
                            local_config = {}
                            local_config_str = ""
                            continue

                        # Set the image path (append in case any others were in the config)
                        img_path = Path(file).parent / Path(m["img_path"].strip())
                        local_config["img_paths"] = local_config.get("img_paths", []) + [str(img_path.resolve())]

                        # Set other config defaults if not supplied
                        local_config["working_dir"] = (
                            Path(local_config["working_dir"]) if "working_dir" in local_config else Path(file).parent
                        )
                        local_config["source_type"] = local_config.get("source_type", "search")
                        local_config["source"] = Path(local_config["source"]) if "source" in local_config else file

                        local_config = self._merge_local_class_attrs(local_config)

                        # Validate the config we have via the schema
                        try:
                            validate_config(self.config_schema, {"outputs": [local_config]}, file_rel_fn, line_number)
                        except ValidationError as e:
                            log.error(e)
                            local_config_str = ""
                            local_config = {}
                            num_errors += 1
                            continue

                        log.debug(
                            "Found markdown {0}, line {1}: {3}{2}{3}{4}".format(
                                img_type,
                                line_number,
                                local_config.get("command", ""),
                                "'" if local_config.get("command") else "",
                                local_config_logmsg,
                            )
                        )
                        img_obj = rich_img.RichImg(**local_config)

                        # Save the image object
                        self.rich_imgs.append(img_obj)

                        # Clear local config
                        local_config = {}
                        local_config_str = ""

                        continue

                    # Look for a local config
                    config_match = config_comment_re.match(line)
                    if config_match:
                        m = config_match.groupdict()

                        # If we don't end the comment on this line, must be a snippet
                        if "-->" not in line:
                            in_config = True

                        # Save config
                        local_config_str = m.get("config_str", "").split("-->")[0] + "\n"

        if num_commands > 0:
            log.info(f"Search: Found {num_commands} commands")
        if num_snippets > 0:
            log.info(f"Search: Found {num_snippets} snippets")
        return num_errors

    def parse_configs(self):
        """Loop through rich-codex config files to send for parsing."""
        configs = []
        for config_fn in self.configs:
            config = Path(config_fn)
            if config.exists():
                log.debug(f"Found config '{config_fn}'")
                configs.append(config)
            else:
                log.debug(f"[dim]Couldn't find '{config_fn}'")

        if len(configs) > 0:
            log.info(f"Found {len(configs)} config file{'s' if len(configs) > 1 else ''}")
        for config in configs:
            with config.open() as fh:
                self.parse_config(config_fn, yaml.safe_load(fh))

    def parse_config(self, config_fn, config):
        """Parse a single rich-codex config file."""
        validate_config(self.config_schema, config, config_fn)

        # Overwrite class-level configs
        for cls in self.class_config_attrs:
            if cls in config:
                setattr(self, cls, config["cls"])

        for output in config["outputs"]:
            log.debug(f"Found valid output in '{config_fn}': {output}")
            output["img_paths"] = [str(Path(img_path_str.strip()).resolve()) for img_path_str in output["img_paths"]]
            output["source_type"] = "config"
            output["source"] = config_fn
            local_config = self._merge_local_class_attrs(output)
            self.rich_imgs.append(rich_img.RichImg(**local_config))

    def collapse_duplicates(self):
        """Collapse duplicate commands."""
        # Remove exact duplicates
        dedup_imgs = set(self.rich_imgs)
        # Merge dups that are the same except for output filename
        merged_imgs = {}
        for ri in dedup_imgs:
            ri_hash = ri._hash_no_fn()
            if ri_hash in merged_imgs:
                merged_imgs[ri_hash].img_paths.extend(ri.img_paths)
            else:
                merged_imgs[ri_hash] = ri
        log.debug(f"Collapsing {len(self.rich_imgs)} image requests to {len(merged_imgs)} deduplicated")
        self.rich_imgs = sorted(merged_imgs.values(), key=lambda x: str(x.command).lower())

    def confirm_commands(self):
        """Prompt the user to confirm running the commands."""
        table = Table(
            title_style="blue",
            title_justify="left",
            box=box.ROUNDED,
            safe_box=True,
            header_style="bold blue",
            border_style="blue",
            row_styles=["green on grey3", "magenta on grey15"],
        )
        table.add_column("Commands to run:")
        table.add_column("Source", justify="right")
        for img_obj in self.rich_imgs:
            if img_obj.command is not None:
                rel_source = Path(img_obj.source).relative_to(Path.cwd())
                source = f" [grey42][link=file:{Path(img_obj.source).absolute()}]{rel_source}[/][/]"
                table.add_row(img_obj.command, source)

        if table.row_count == 0:
            return True

        self.console.print(table)

        if self.no_confirm:
            return True

        confirm = Prompt.ask(
            "Do you want to run these commands? (All / Some / None)", choices=["a", "s", "n"], console=self.console
        )
        if confirm == "a":
            log.info("Running all commands")
            return True
        elif confirm == "n":
            log.info("Skipping all outputs that require running a command")
            self.rich_imgs = [ri for ri in self.rich_imgs if ri.command is None]
            return False
        else:
            log.info("Please select commands individually")
            self.rich_imgs = [ri for ri in self.rich_imgs if ri.confirm_command()]
            return None

    def check_duplicate_paths(self):
        """Check that we don't have any duplicate output file paths."""
        img_paths_src = {}
        for ri in self.rich_imgs:
            for img_path in ri.img_paths:
                if img_path in img_paths_src:
                    img_paths_src[img_path].append(ri.source)
                else:
                    img_paths_src[img_path] = [ri.source]
        for img_path, src in img_paths_src.items():
            if len(src) > 1:
                img_path_rel = Path(img_path).relative_to(Path.cwd())
                src_paths = "', '".join(set(str(s.relative_to(Path.cwd())) for s in src))
                log.warning(f"Duplicate output file path '{img_path_rel}' found in '{src_paths}'")

    def save_all_images(self):
        """Save the images that we have collected."""
        for img_obj in self.rich_imgs:
            img_obj.get_output()
            img_obj.save_images()
            self.saved_img_paths += img_obj.saved_img_paths
            self.num_img_saved += img_obj.num_img_saved
            self.num_img_skipped += img_obj.num_img_skipped



from setuptools import setup

setup(
    name="rich-codex",
    install_requires=[
        "rich>=12.4.3",
        "rich-click>=1.5",
        "rich-cli",  # For convience only
        "importlib-metadata; python_version < '3.8'",
        "pyyaml",
        "levenshtein>=0.18.1",
        "jsonschema",
        "GitPython",
    ],
    extras_require={
        "cairo": "CairoSVG>=2.5.2",
        "dev": "pre-commit",
    },
)
